from __future__ import annotations

import random
from typing import Any

import pytest

from fabrix.errors import LLMOutputError
from fabrix.llm.oauth_codex import OAuthCodexStateProvider

_FORBIDDEN = {"$ref", "$defs", "definitions", "oneOf", "const"}


def _assert_no_keywords(node: Any) -> None:
    if isinstance(node, dict):
        for key, value in node.items():
            assert key not in _FORBIDDEN
            _assert_no_keywords(value)
        return

    if isinstance(node, list):
        for item in node:
            _assert_no_keywords(item)


def _assert_object_consistency(node: Any) -> None:
    if isinstance(node, dict):
        if node.get("type") == "object":
            properties = node.get("properties")
            required = node.get("required")
            assert isinstance(properties, dict)
            assert isinstance(required, list)
            assert required == list(properties.keys())
            assert node.get("additionalProperties") is False

        for value in node.values():
            _assert_object_consistency(value)
        return

    if isinstance(node, list):
        for item in node:
            _assert_object_consistency(item)


def _random_literal(rng: random.Random) -> Any:
    pool: list[Any] = [None, True, False, rng.randint(-10, 10), f"v{rng.randint(0, 9)}"]
    return rng.choice(pool)


def _random_schema_node(rng: random.Random, *, depth: int, def_names: list[str]) -> dict[str, Any]:
    if depth <= 0:
        leaf_options = [
            {"type": "string"},
            {"type": "integer"},
            {"const": _random_literal(rng)},
            {
                "oneOf": [
                    {"type": "string"},
                    {"const": _random_literal(rng)},
                ]
            },
        ]
        return rng.choice(leaf_options)

    kind = rng.choice(["object", "array", "oneOf", "const", "primitive", "ref"])

    if kind == "object":
        prop_count = rng.randint(0, 3)
        properties: dict[str, Any] = {}
        for idx in range(prop_count):
            properties[f"p{idx}"] = _random_schema_node(rng, depth=depth - 1, def_names=def_names)
        return {
            "type": "object",
            "properties": properties,
            "required": list(reversed(list(properties.keys()))),
            "additionalProperties": True,
        }

    if kind == "array":
        return {
            "type": "array",
            "items": _random_schema_node(rng, depth=depth - 1, def_names=def_names),
        }

    if kind == "oneOf":
        return {
            "oneOf": [
                _random_schema_node(rng, depth=depth - 1, def_names=def_names),
                _random_schema_node(rng, depth=depth - 1, def_names=def_names),
            ]
        }

    if kind == "const":
        return {"const": _random_literal(rng)}

    if kind == "ref" and def_names:
        return {"$ref": f"#/$defs/{rng.choice(def_names)}"}

    return rng.choice([{"type": "string"}, {"type": "number"}, {"type": "boolean"}])


def _build_random_schema(seed: int) -> dict[str, Any]:
    rng = random.Random(seed)

    def_names = [f"Def{i}" for i in range(rng.randint(1, 4))]
    defs: dict[str, Any] = {}
    for name in def_names:
        def_schema = _random_schema_node(rng, depth=2, def_names=[])
        if def_schema.get("type") != "object":
            def_schema = {
                "type": "object",
                "properties": {"value": def_schema},
                "required": [],
                "additionalProperties": True,
            }
        defs[name] = def_schema

    prop_count = rng.randint(1, 4)
    properties: dict[str, Any] = {}
    for idx in range(prop_count):
        properties[f"field{idx}"] = _random_schema_node(rng, depth=3, def_names=def_names)

    return {
        "type": "object",
        "properties": properties,
        "required": [],
        "additionalProperties": True,
        "$defs": defs,
    }


def test_normalization_fuzz_outputs_provider_safe_schema(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)

    for seed in range(50):
        schema = _build_random_schema(seed)
        normalized = provider._normalize_schema(schema)

        _assert_no_keywords(normalized)
        _assert_object_consistency(normalized)


def test_normalization_rejects_unresolvable_ref(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)

    schema = {
        "type": "object",
        "properties": {"x": {"$ref": "#/$defs/Missing"}},
        "required": ["x"],
        "additionalProperties": False,
    }

    with pytest.raises(LLMOutputError, match="unresolvable \\$ref pointer"):
        provider._normalize_schema(schema)


def test_normalization_rejects_non_local_ref(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)

    schema = {
        "type": "object",
        "properties": {"x": {"$ref": "https://example.com/schema.json#/x"}},
        "required": ["x"],
        "additionalProperties": False,
    }

    with pytest.raises(LLMOutputError, match="unsupported \\$ref pointer"):
        provider._normalize_schema(schema)


def test_normalization_rejects_circular_refs(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)

    schema = {
        "$defs": {
            "A": {"$ref": "#/$defs/B"},
            "B": {"$ref": "#/$defs/A"},
        },
        "type": "object",
        "properties": {
            "root": {"$ref": "#/$defs/A"},
        },
        "required": ["root"],
        "additionalProperties": False,
    }

    with pytest.raises(LLMOutputError, match="circular \\$ref detected"):
        provider._normalize_schema(schema)


def test_normalization_rejects_type_union_array_without_items(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)

    schema = {
        "type": "object",
        "properties": {
            "value": {
                "type": ["string", "array"],
            }
        },
        "required": ["value"],
        "additionalProperties": False,
    }

    with pytest.raises(LLMOutputError, match="array schema missing items"):
        provider._normalize_schema(schema)



from __future__ import annotations

import asyncio
import sys
from pathlib import Path

from pydantic import BaseModel, Field

EXAMPLES_ROOT = Path(__file__).resolve().parents[1]
if str(EXAMPLES_ROOT) not in sys.path:
    sys.path.insert(0, str(EXAMPLES_ROOT))

from common import run_case
from fabrix import Agent
from fabrix.messages import TextMessage
from fabrix.tools import ToolOutput


class AddInput(BaseModel):
    a: int = Field(ge=-10_000, le=10_000)
    b: int = Field(ge=-10_000, le=10_000)


def add_numbers(payload: AddInput) -> ToolOutput:
    """Add two integers and return the sum."""
    return ToolOutput.json({"sum": payload.a + payload.b})


async def main() -> None:
    agent = Agent(
        instructions=(
            "Solve the task accurately. Prefer tool usage for arithmetic tasks. "
            "When done, return response state with next_state set to null."
        ),
        state_models={"reasoning": "gpt-5.3-codex"},
        tools=[add_numbers],
    )

    messages = [TextMessage(role="user", text="Use add_numbers to compute 38 + 4 and explain briefly.")]
    await run_case(agent, messages=messages)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import asyncio
import json
import sys
from typing import Literal
from pathlib import Path

from pydantic import BaseModel, Field

EXAMPLES_ROOT = Path(__file__).resolve().parents[1]
if str(EXAMPLES_ROOT) not in sys.path:
    sys.path.insert(0, str(EXAMPLES_ROOT))

from common import print_event
from fabrix import Agent
from fabrix.messages import TextMessage
from fabrix.tools import ToolOutput

RiskLevel = Literal["low", "medium", "high"]
LoadLevel = Literal["low", "medium", "high"]

RISK_RANK: dict[RiskLevel, int] = {"low": 0, "medium": 1, "high": 2}
COMPLIANCE_SCORE: dict[RiskLevel, float] = {"low": 95.0, "medium": 70.0, "high": 35.0}

PROJECT_PLANS = [
    {
        "plan_id": "alpha-rehost",
        "infra_cost_k": 210,
        "launch_weeks": 6,
        "reliability_gain": 72,
        "compliance_risk": "low",
        "engineering_load": "high",
    },
    {
        "plan_id": "beta-hybrid",
        "infra_cost_k": 170,
        "launch_weeks": 8,
        "reliability_gain": 58,
        "compliance_risk": "medium",
        "engineering_load": "medium",
    },
    {
        "plan_id": "gamma-refactor",
        "infra_cost_k": 250,
        "launch_weeks": 12,
        "reliability_gain": 86,
        "compliance_risk": "low",
        "engineering_load": "high",
    },
    {
        "plan_id": "delta-minimal",
        "infra_cost_k": 120,
        "launch_weeks": 4,
        "reliability_gain": 40,
        "compliance_risk": "high",
        "engineering_load": "low",
    },
]

HARD_CONSTRAINTS = {
    "max_cost_k": 220,
    "max_launch_weeks": 8,
    "max_compliance_risk": "medium",
}

BASE_WEIGHTS = {
    "reliability": 0.45,
    "cost": 0.2,
    "delivery": 0.2,
    "compliance": 0.15,
}

SOFT_OBJECTIVES = {
    "prefer_lower_engineering_load_when_scores_close": True,
    "target_min_reliability_gain": 55,
    "favor_schedule_headroom_for_peak_season_readiness": True,
    "if_winner_is_unstable_under_stress_test": "include fallback plan",
}


class Plan(BaseModel):
    plan_id: str = Field(min_length=1)
    infra_cost_k: int = Field(ge=0)
    launch_weeks: int = Field(ge=1)
    reliability_gain: int = Field(ge=0, le=100)
    compliance_risk: RiskLevel
    engineering_load: LoadLevel


class Constraints(BaseModel):
    max_cost_k: int = Field(ge=1)
    max_launch_weeks: int = Field(ge=1)
    max_compliance_risk: RiskLevel


class FilterFeasibleInput(BaseModel):
    plans: list[Plan] = Field(min_length=1)
    constraints: Constraints


def filter_feasible_plans(payload: FilterFeasibleInput) -> ToolOutput:
    feasible: list[dict[str, int | str]] = []
    risk_limit = RISK_RANK[payload.constraints.max_compliance_risk]

    for plan in payload.plans:
        if plan.infra_cost_k > payload.constraints.max_cost_k:
            continue
        if plan.launch_weeks > payload.constraints.max_launch_weeks:
            continue
        if RISK_RANK[plan.compliance_risk] > risk_limit:
            continue

        feasible.append(
            {
                "plan_id": plan.plan_id,
                "infra_cost_k": plan.infra_cost_k,
                "launch_weeks": plan.launch_weeks,
                "reliability_gain": plan.reliability_gain,
                "compliance_risk": plan.compliance_risk,
                "engineering_load": plan.engineering_load,
                "cost_headroom_k": payload.constraints.max_cost_k - plan.infra_cost_k,
                "schedule_headroom_weeks": payload.constraints.max_launch_weeks - plan.launch_weeks,
            }
        )

    return ToolOutput.json(feasible)


class Weights(BaseModel):
    reliability: float = Field(gt=0)
    cost: float = Field(gt=0)
    delivery: float = Field(gt=0)
    compliance: float = Field(gt=0)


class FeasiblePlan(BaseModel):
    plan_id: str = Field(min_length=1)
    infra_cost_k: int = Field(ge=0)
    launch_weeks: int = Field(ge=1)
    reliability_gain: int = Field(ge=0, le=100)
    compliance_risk: RiskLevel
    engineering_load: LoadLevel
    cost_headroom_k: int
    schedule_headroom_weeks: int


class RankPlansInput(BaseModel):
    feasible_plans: list[FeasiblePlan] = Field(min_length=1)
    weights: Weights


def _weighted_score(
    *,
    infra_cost_k: int,
    launch_weeks: int,
    reliability_gain: int,
    compliance_risk: RiskLevel,
    weights: Weights,
) -> float:
    cost_score = max(0.0, min(100.0, 100.0 - float(infra_cost_k) * 0.35))
    delivery_score = max(0.0, min(100.0, 100.0 - float(launch_weeks) * 8.0))
    reliability_score = float(reliability_gain)
    compliance_score = COMPLIANCE_SCORE[compliance_risk]
    return (
        reliability_score * weights.reliability
        + cost_score * weights.cost
        + delivery_score * weights.delivery
        + compliance_score * weights.compliance
    )


def _score_plan(
    *,
    plan: FeasiblePlan,
    weights: Weights,
) -> float:
    return _weighted_score(
        infra_cost_k=plan.infra_cost_k,
        launch_weeks=plan.launch_weeks,
        reliability_gain=plan.reliability_gain,
        compliance_risk=plan.compliance_risk,
        weights=weights,
    )


def rank_feasible_plans(payload: RankPlansInput) -> ToolOutput:
    ranked: list[dict[str, int | float | str]] = []

    for plan in payload.feasible_plans:
        total_score = round(_score_plan(plan=plan, weights=payload.weights), 2)
        ranked.append(
            {
                "plan_id": plan.plan_id,
                "score": total_score,
                "reliability_gain": plan.reliability_gain,
                "infra_cost_k": plan.infra_cost_k,
                "launch_weeks": plan.launch_weeks,
                "compliance_risk": plan.compliance_risk,
                "engineering_load": plan.engineering_load,
                "cost_headroom_k": plan.cost_headroom_k,
                "schedule_headroom_weeks": plan.schedule_headroom_weeks,
            }
        )

    ranked.sort(key=lambda row: (-float(row["score"]), int(row["infra_cost_k"]), str(row["plan_id"])))
    return ToolOutput.json(ranked)


class RankedPlan(BaseModel):
    plan_id: str = Field(min_length=1)
    score: float
    reliability_gain: int = Field(ge=0, le=100)
    infra_cost_k: int = Field(ge=0)
    launch_weeks: int = Field(ge=1)
    compliance_risk: RiskLevel
    engineering_load: LoadLevel
    cost_headroom_k: int
    schedule_headroom_weeks: int


class StressTestInput(BaseModel):
    ranked_plans: list[RankedPlan] = Field(min_length=1)


def stress_test_top_plans(payload: StressTestInput) -> ToolOutput:
    top = payload.ranked_plans[:2]
    if not top:
        return ToolOutput.json({"scenarios": []})

    scenarios = {
        "reliability_heavy": Weights(reliability=0.6, cost=0.15, delivery=0.1, compliance=0.15),
        "cost_crunch": Weights(reliability=0.3, cost=0.4, delivery=0.2, compliance=0.1),
        "delivery_urgent": Weights(reliability=0.35, cost=0.15, delivery=0.4, compliance=0.1),
    }

    results: list[dict[str, str | float]] = []
    for scenario_name, weights in scenarios.items():
        winner = max(
            top,
            key=lambda plan: _weighted_score(
                infra_cost_k=plan.infra_cost_k,
                launch_weeks=plan.launch_weeks,
                reliability_gain=plan.reliability_gain,
                compliance_risk=plan.compliance_risk,
                weights=weights,
            ),
        )
        results.append(
            {
                "scenario": scenario_name,
                "winner_plan_id": winner.plan_id,
                "winner_score": round(
                    _weighted_score(
                        infra_cost_k=winner.infra_cost_k,
                        launch_weeks=winner.launch_weeks,
                        reliability_gain=winner.reliability_gain,
                        compliance_risk=winner.compliance_risk,
                        weights=weights,
                    ),
                    2,
                ),
            }
        )

    stability = len({item["winner_plan_id"] for item in results}) == 1
    return ToolOutput.json({"scenarios": results, "stable_winner": stability})


class ScenarioResult(BaseModel):
    scenario: str = Field(min_length=1)
    winner_plan_id: str = Field(min_length=1)
    winner_score: float


class RenderBriefInput(BaseModel):
    ranked_plans: list[RankedPlan] = Field(min_length=1)
    stress_scenarios: list[ScenarioResult] = Field(min_length=1)
    stable_winner: bool


def render_decision_brief(payload: RenderBriefInput) -> ToolOutput:
    top = payload.ranked_plans[0]
    scenario_winners = ", ".join(
        f"{row.scenario}:{row.winner_plan_id}" for row in payload.stress_scenarios
    )
    stability_note = "stable across scenarios" if payload.stable_winner else "sensitive to weights"
    brief = (
        f"Recommend `{top.plan_id}` (score={top.score}). "
        f"Trade-offs: cost={top.infra_cost_k}k, launch={top.launch_weeks}w, "
        f"reliability_gain={top.reliability_gain}, compliance={top.compliance_risk}. "
        f"Stress-test winners -> {scenario_winners}. "
        f"Decision robustness: {stability_note}."
    )
    return ToolOutput.text(brief)


async def main() -> None:
    agent = Agent(
        instructions=(
            "You are a principal architect making a multi-constraint migration decision. "
            "Use the registered tools to evaluate feasibility, rank candidates, and validate robustness. "
            "A good flow usually starts by filtering feasibility, then ranking feasible candidates, "
            "then stress-testing likely winners, and finally rendering the decision brief. "
            "Before finalizing, make sure the recommendation accounts for hard constraints and soft objectives, "
            "including what to do if stress-test winners are unstable. "
            "Do not repeat the same tool call with identical arguments. "
            "When done, set response to render_decision_brief output and set next_state to null."
        ),
        state_models={"reasoning": "gpt-5.3-codex"},
        tools=[
            filter_feasible_plans,
            rank_feasible_plans,
            stress_test_top_plans,
            render_decision_brief,
        ],
    )

    plans_json = json.dumps(PROJECT_PLANS, ensure_ascii=True)
    constraints_json = json.dumps(HARD_CONSTRAINTS, ensure_ascii=True)
    weights_json = json.dumps(BASE_WEIGHTS, ensure_ascii=True)
    objectives_json = json.dumps(SOFT_OBJECTIVES, ensure_ascii=True)

    messages = [
        TextMessage(
            role="user",
            text=(
                "Select the best migration plan with transparent trade-offs. "
                f"plans={plans_json} constraints={constraints_json} "
                f"weights={weights_json} soft_objectives={objectives_json}"
            ),
        )
    ]

    step_flow: list[str] = []
    last_step_event: tuple[int, str] | None = None
    current_reasoning_streak = 0
    max_reasoning_streak = 0
    initial_reasoning_streak = 0
    in_initial_reasoning_block = True

    async for event in agent.run_stream(messages=messages):
        print_event(event)
        step_event = (event.step, event.event_type)
        if step_event == last_step_event:
            continue

        last_step_event = step_event
        step_flow.append(event.event_type)

        if event.event_type == "reasoning":
            current_reasoning_streak += 1
            max_reasoning_streak = max(max_reasoning_streak, current_reasoning_streak)
            if in_initial_reasoning_block:
                initial_reasoning_streak += 1
            continue

        current_reasoning_streak = 0
        in_initial_reasoning_block = False

    print(f"\nstep_flow={' -> '.join(step_flow)}")
    print(f"initial_reasoning_streak={initial_reasoning_streak}")
    print(f"max_reasoning_streak={max_reasoning_streak}")


if __name__ == "__main__":
    asyncio.run(main())



from .state import (
    AgentState,
    BaseState,
    NextState,
    PlannedToolCall,
    ReasoningState,
    ResponseState,
    StateEnvelope,
    ToolCallState,
)
from .transitions import ALLOWED_TRANSITIONS, allowed_next_states, validate_transition

__all__ = [
    "NextState",
    "BaseState",
    "PlannedToolCall",
    "ReasoningState",
    "ToolCallState",
    "ResponseState",
    "AgentState",
    "StateEnvelope",
    "ALLOWED_TRANSITIONS",
    "allowed_next_states",
    "validate_transition",
]



from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest
from pydantic import BaseModel

import fabrix.llm.oauth_codex as oauth_codex_module
from fabrix.agent import Agent
from fabrix.errors import LLMOutputError
from fabrix.graph.state import MAX_TOOL_CALLS_PER_STEP, NextState
from fabrix.graph.transitions import allowed_next_states
from fabrix.llm.oauth_codex import OAuthCodexStateProvider
from fabrix.messages import ImageMessage, TextMessage
from fabrix.tools.registry import ToolRegistry


class AddInput(BaseModel):
    a: int
    b: int


def add_numbers(payload: AddInput) -> int:
    return payload.a + payload.b


class RecursiveInput(BaseModel):
    value: int
    child: RecursiveInput | None = None


RecursiveInput.model_rebuild()


def consume_recursive(payload: RecursiveInput) -> int:
    return payload.value


def _assert_no_keywords(node: Any, keywords: set[str]) -> None:
    if isinstance(node, dict):
        for key, value in node.items():
            assert key not in keywords
            _assert_no_keywords(value, keywords)
        return

    if isinstance(node, list):
        for item in node:
            _assert_no_keywords(item, keywords)


def _find_message_with_text(messages: list[dict[str, Any]], needle: str) -> dict[str, Any]:
    for message in messages:
        content = message.get("content")
        if isinstance(content, str) and needle in content:
            return message
        if not isinstance(content, list):
            continue
        for part in content:
            if not isinstance(part, dict):
                continue
            text = part.get("text")
            if isinstance(text, str) and needle in text:
                return message
    raise AssertionError(f"message containing `{needle}` not found")


class _LooseJSONStringClient:
    def __init__(self) -> None:
        self.calls: list[dict[str, Any]] = []

    async def agenerate(self, **kwargs: Any) -> Any:
        self.calls.append(kwargs)
        return (
            "Here is the result.\n"
            "```json\n"
            '{"state":{"state_type":"reasoning","next_state":"response","reasoning":"done","focus":"finalize"}}\n'
            "```\n"
            "Use it."
        )


def test_tool_call_items_use_anyof_and_name_arguments_only(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    tool_schemas = ToolRegistry.from_callables([add_numbers]).schemas()

    schema = provider._build_output_schema(
        current_state=NextState.tool_call,
        tool_schemas=tool_schemas,
    )

    any_of = schema["json_schema"]["schema"]["properties"]["state"]["properties"]["tool_calls"][
        "items"
    ]["anyOf"]
    assert len(any_of) == 1

    item_schema = any_of[0]
    assert set(item_schema["properties"]) == {"name", "arguments"}
    assert item_schema["required"] == ["name", "arguments"]


def test_state_schema_preserves_pydantic_min_length_constraints(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)

    reasoning_schema = provider._build_output_schema(
        current_state=NextState.reasoning,
        tool_schemas=[],
    )
    reasoning_props = reasoning_schema["json_schema"]["schema"]["properties"]["state"]["properties"]
    assert reasoning_props["reasoning"]["minLength"] == 1
    assert reasoning_props["focus"]["minLength"] == 1

    response_schema = provider._build_output_schema(
        current_state=NextState.response,
        tool_schemas=[],
    )
    response_props = response_schema["json_schema"]["schema"]["properties"]["state"]["properties"]
    response_any_of = response_props["response"]["anyOf"]
    assert any(item.get("type") == "null" for item in response_any_of if isinstance(item, dict))
    assert any(item.get("type") == "string" for item in response_any_of if isinstance(item, dict))
    assert "parts" in response_props


def test_tool_call_schema_has_max_items_constraint(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    tool_schemas = ToolRegistry.from_callables([add_numbers]).schemas()

    schema = provider._build_output_schema(
        current_state=NextState.tool_call,
        tool_schemas=tool_schemas,
    )
    tool_calls_schema = schema["json_schema"]["schema"]["properties"]["state"]["properties"][
        "tool_calls"
    ]
    assert tool_calls_schema["minItems"] == 1
    assert tool_calls_schema["maxItems"] == MAX_TOOL_CALLS_PER_STEP


def test_response_schema_json_part_data_uses_scalar_anyof(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    response_schema = provider._build_output_schema(
        current_state=NextState.response,
        tool_schemas=[],
    )
    response_props = response_schema["json_schema"]["schema"]["properties"]["state"]["properties"]
    parts_any_of = response_props["parts"]["anyOf"]
    array_schema = next(
        item for item in parts_any_of if isinstance(item, dict) and item.get("type") == "array"
    )
    item_any_of = array_schema["items"]["anyOf"]
    json_part_schema = next(
        item
        for item in item_any_of
        if isinstance(item, dict)
        and item.get("properties", {}).get("type", {}).get("enum") == ["json"]
    )
    data_schema = json_part_schema["properties"]["data"]
    any_of = data_schema.get("anyOf")
    assert isinstance(any_of, list)
    types = {item.get("type") for item in any_of if isinstance(item, dict)}
    assert types == {"string", "number", "integer", "boolean", "null"}
    assert "array" not in types
    assert "object" not in types


def test_output_schema_has_no_unsupported_keywords_for_all_states(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    tool_schemas = ToolRegistry.from_callables([add_numbers]).schemas()
    forbidden = {"$ref", "$defs", "definitions", "oneOf", "const"}

    for state in NextState:
        if state is NextState.tool_call:
            schema = provider._build_output_schema(current_state=state, tool_schemas=tool_schemas)
        else:
            schema = provider._build_output_schema(current_state=state, tool_schemas=[])
        _assert_no_keywords(schema, forbidden)


@pytest.mark.asyncio
async def test_provider_parses_fenced_json_from_string_payload() -> None:
    client = _LooseJSONStringClient()
    provider = OAuthCodexStateProvider(instructions="x", client=client)

    envelope = await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[],
        current_state=NextState.reasoning,
        step=1,
        tool_schemas=[],
    )

    assert len(client.calls) == 1
    assert client.calls[0]["output_schema"] is not None
    assert client.calls[0]["strict_output"] is True
    assert envelope.state.state_type == "reasoning"
    assert envelope.state.next_state == NextState.response
    assert envelope.state.reasoning == "done"
    assert envelope.state.focus == "finalize"


@pytest.mark.asyncio
async def test_provider_uses_dynamic_output_schema(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    tool_schemas = ToolRegistry.from_callables([add_numbers]).schemas()

    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[],
        current_state=NextState.reasoning,
        step=1,
        tool_schemas=tool_schemas,
    )

    output_schema = fake_client.calls[-1]["output_schema"]
    assert isinstance(output_schema, dict)
    assert output_schema["type"] == "json_schema"

    state_schema = output_schema["json_schema"]["schema"]["properties"]["state"]
    state_type_schema = state_schema["properties"]["state_type"]
    assert state_type_schema.get("enum") == ["reasoning"]
    assert state_schema["properties"]["next_state"]["enum"] == [
        state.value for state in allowed_next_states(NextState.reasoning)
    ]

    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[],
        current_state=NextState.reasoning,
        step=2,
        tool_schemas=[],
    )
    no_tool_output_schema = fake_client.calls[-1]["output_schema"]
    no_tool_state_schema = no_tool_output_schema["json_schema"]["schema"]["properties"]["state"]
    assert no_tool_state_schema["properties"]["next_state"]["enum"] == [
        state.value for state in allowed_next_states(NextState.reasoning)
    ]


def test_response_schema_next_state_is_nullable(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    schema = provider._build_output_schema(current_state=NextState.response, tool_schemas=[])
    next_state_schema = schema["json_schema"]["schema"]["properties"]["state"]["properties"]["next_state"]
    any_of = next_state_schema.get("anyOf")
    assert isinstance(any_of, list)
    assert any(item.get("type") == "null" for item in any_of if isinstance(item, dict))
    non_null = [
        item for item in any_of if isinstance(item, dict) and item.get("type") != "null"
    ]
    assert len(non_null) == 1
    assert non_null[0].get("enum") == [state.value for state in allowed_next_states(NextState.response)]


@pytest.mark.asyncio
async def test_provider_passes_messages_to_client_in_order(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    messages = [
        TextMessage(role="user", text="Describe this image"),
        ImageMessage(
            role="user",
            image="https://example.com/image.png",
            text="Focus on warnings",
        ),
        TextMessage(role="user", text="Focus on warnings"),
    ]

    await provider.generate_state(
        messages=messages,
        history=[],
        current_state=NextState.reasoning,
        step=1,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    assert sent_messages[0]["role"] == "system"
    assert isinstance(sent_messages[0]["content"], str)
    assert [item["role"] for item in sent_messages[1:4]] == ["user", "user", "user"]
    assert sent_messages[1]["content"] == "Describe this image"
    content = sent_messages[2]["content"]
    assert isinstance(content, list)
    assert [part["type"] for part in content] == ["input_text", "input_image"]
    assert sent_messages[3]["content"] == "Focus on warnings"


@pytest.mark.asyncio
async def test_provider_appends_multimodal_history_image_parts(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[
            {
                "kind": "tool_result",
                "step": 2,
                "tool_name": "vision_tool",
                "call_id": "c2",
                "ok": True,
                "output": {
                    "parts": [
                        {
                            "type": "image",
                            "image_url": "https://example.com/tool.png",
                            "caption": "detected object",
                        }
                    ]
                },
            }
        ],
        current_state=NextState.reasoning,
        step=3,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    tool_result_message = _find_message_with_text(sent_messages, "tool_result step=2 tool=vision_tool")
    content = tool_result_message["content"]
    assert isinstance(content, list)
    assert [part["type"] for part in content] == ["input_text", "input_text", "input_image"]
    assert content[-1]["image_url"] == "https://example.com/tool.png"


@pytest.mark.asyncio
async def test_provider_converts_local_image_history_part_to_data_url(
    fake_client: Any, tmp_path: Path
) -> None:
    image_file = tmp_path / "tool.png"
    image_file.write_bytes(b"\x89PNG\r\n\x1a\n\x00")

    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[
            {
                "kind": "tool_result",
                "step": 2,
                "tool_name": "vision_tool",
                "call_id": "c2",
                "ok": True,
                "output": {
                    "parts": [
                        {
                            "type": "image",
                            "image_url": str(image_file),
                        }
                    ]
                },
            }
        ],
        current_state=NextState.reasoning,
        step=3,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    tool_result_message = _find_message_with_text(sent_messages, "tool_result step=2 tool=vision_tool")
    content = tool_result_message["content"]
    assert isinstance(content, list)
    assert [part["type"] for part in content] == ["input_text", "input_image"]
    assert content[-1]["image_url"].startswith("data:image/png;base64,")


@pytest.mark.asyncio
async def test_provider_skips_missing_local_image_history_part_with_warning(
    fake_client: Any, tmp_path: Path
) -> None:
    missing_path = tmp_path / "missing.png"

    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[
            {
                "kind": "tool_result",
                "step": 2,
                "tool_name": "vision_tool",
                "call_id": "c2",
                "ok": True,
                "output": {
                    "parts": [
                        {
                            "type": "image",
                            "image_url": str(missing_path),
                            "caption": "detected object",
                        }
                    ]
                },
            }
        ],
        current_state=NextState.reasoning,
        step=3,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    tool_result_message = _find_message_with_text(sent_messages, "tool_result step=2 tool=vision_tool")
    content = tool_result_message["content"]
    assert isinstance(content, list)
    assert [part["type"] for part in content] == ["input_text", "input_text", "input_text"]
    assert content[1]["text"] == "detected object"
    assert "tool_result image skipped:" in content[2]["text"]
    assert not any(part["type"] == "input_image" for part in content)


@pytest.mark.asyncio
async def test_provider_appends_multimodal_history_text_and_json_parts(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[
            {
                "kind": "tool_result",
                "step": 4,
                "tool_name": "formatter",
                "call_id": "c4",
                "ok": True,
                "output": {
                    "parts": [
                        {"type": "text", "text": "formatted"},
                        {"type": "json", "data": {"a": 1}},
                    ]
                },
            }
        ],
        current_state=NextState.reasoning,
        step=5,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    tool_result_message = _find_message_with_text(sent_messages, "tool_result step=4 tool=formatter")
    content = tool_result_message["content"]
    assert isinstance(content, list)
    assert content[0]["type"] == "input_text"
    assert "tool_result step=4" in content[0]["text"]
    assert any(part["type"] == "input_text" and part["text"] == "formatted" for part in content)
    assert any(part["type"] == "input_text" and part["text"] == '{"a":1}' for part in content)


@pytest.mark.asyncio
async def test_provider_appends_failed_tool_history_as_single_text(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[
            {
                "kind": "tool_result",
                "step": 6,
                "tool_name": "broken",
                "call_id": "c6",
                "ok": False,
                "error": "boom",
            }
        ],
        current_state=NextState.reasoning,
        step=7,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    tool_result_message = _find_message_with_text(sent_messages, "tool_result step=6 tool=broken")
    content = tool_result_message["content"]
    assert isinstance(content, list)
    assert len(content) == 1
    assert content[0]["type"] == "input_text"
    assert "ok=False" in content[0]["text"]
    assert "error=boom" in content[0]["text"]


@pytest.mark.asyncio
async def test_provider_serializes_tool_call_history_with_tool_results(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[
            {
                "kind": "tool_call",
                "step": 2,
                "next_state": "reasoning",
                "tool_calls": [
                    {
                        "name": "vision_tool",
                        "call_id": "s2_c1",
                        "arguments": {"target": "error"},
                        "why": "inspect image",
                    }
                ],
                "tool_results": [
                    {
                        "step": 2,
                        "tool_name": "vision_tool",
                        "call_id": "s2_c1",
                        "ok": True,
                        "output": {"parts": [{"type": "text", "text": "detected"}]},
                        "error": None,
                    }
                ],
            }
        ],
        current_state=NextState.reasoning,
        step=3,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    tool_call_summary = _find_message_with_text(sent_messages, "tool_call step=2")
    summary_parts = tool_call_summary["content"]
    assert isinstance(summary_parts, list)
    assert any("planned_tool_call name=vision_tool" in part.get("text", "") for part in summary_parts)
    tool_result = _find_message_with_text(sent_messages, "tool_result step=2 tool=vision_tool")
    tool_result_parts = tool_result["content"]
    assert isinstance(tool_result_parts, list)
    assert any(part.get("text") == "detected" for part in tool_result_parts)


@pytest.mark.asyncio
async def test_provider_serializes_reasoning_and_response_history(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[
            {
                "kind": "reasoning",
                "step": 1,
                "reasoning": "need one more check",
                "focus": "validate result",
                "next_state": "response",
            },
            {
                "kind": "response",
                "step": 2,
                "next_state": "reasoning",
                "response": "intermediate",
                "parts": [
                    {"type": "json", "data": {"a": 1}},
                    {"type": "image", "image_url": "https://example.com/out.png", "caption": "preview"},
                ],
            },
        ],
        current_state=NextState.reasoning,
        step=3,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    reasoning_message = _find_message_with_text(sent_messages, "reasoning step=1")
    assert reasoning_message["role"] == "assistant"
    response_message = _find_message_with_text(sent_messages, "intermediate")
    response_parts = response_message["content"]
    assert isinstance(response_parts, list)
    assert any(part.get("type") == "output_text" and part.get("text") == "preview" for part in response_parts)
    assert any(part.get("type") == "input_image" for part in response_parts)


@pytest.mark.asyncio
async def test_provider_keeps_unknown_history_items_as_raw_text(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    await provider.generate_state(
        messages=[TextMessage(role="user", text="hello")],
        history=[{"kind": "x_custom", "value": {"a": 1}}],
        current_state=NextState.reasoning,
        step=1,
        tool_schemas=[],
    )

    sent_messages = fake_client.calls[-1]["messages"]
    unknown_message = _find_message_with_text(sent_messages, "history_item index=0 raw=")
    content = unknown_message["content"]
    assert isinstance(content, list)
    assert '"kind":"x_custom"' in content[0]["text"]


@pytest.mark.asyncio
async def test_provider_serializes_input_messages_once_per_generate_state(
    fake_client: Any, monkeypatch: pytest.MonkeyPatch
) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    call_count = 0
    original = oauth_codex_module.to_oauth_message

    def wrapped(message: TextMessage | ImageMessage) -> dict[str, Any]:
        nonlocal call_count
        call_count += 1
        return original(message)

    monkeypatch.setattr(oauth_codex_module, "to_oauth_message", wrapped)

    messages = [
        TextMessage(role="user", text="hello"),
        ImageMessage(role="user", image="https://example.com/a.png"),
    ]
    await provider.generate_state(
        messages=messages,
        history=[],
        current_state=NextState.reasoning,
        step=1,
        tool_schemas=[],
    )
    assert call_count == len(messages)


def test_provider_output_schema_cache_returns_deep_copy(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    tool_schemas = ToolRegistry.from_callables([add_numbers]).schemas()

    schema_first = provider._build_output_schema(
        current_state=NextState.tool_call,
        tool_schemas=tool_schemas,
    )
    schema_first["json_schema"]["schema"]["properties"]["state"]["properties"]["tool_calls"][
        "minItems"
    ] = 999

    schema_second = provider._build_output_schema(
        current_state=NextState.tool_call,
        tool_schemas=tool_schemas,
    )
    assert (
        schema_second["json_schema"]["schema"]["properties"]["state"]["properties"]["tool_calls"][
            "minItems"
        ]
        == 1
    )


@pytest.mark.asyncio
async def test_provider_rejects_non_message_objects(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)

    with pytest.raises(TypeError, match="messages must contain TextMessage/ImageMessage objects"):
        await provider.generate_state(
            messages=[{"role": "user", "text": "hello"}],  # type: ignore[list-item]
            history=[],
            current_state=NextState.reasoning,
            step=1,
            tool_schemas=[],
        )


def test_validate_tool_schemas_rejects_recursive_model(fake_client: Any) -> None:
    provider = OAuthCodexStateProvider(instructions="x", client=fake_client)
    tool_schemas = ToolRegistry.from_callables([consume_recursive]).schemas()

    with pytest.raises(LLMOutputError, match="failed to strictify tool parameter schema"):
        provider.validate_tool_schemas(tool_schemas)


def test_agent_preflight_fails_fast_for_incompatible_tool_schema(fake_client: Any) -> None:
    with pytest.raises(LLMOutputError, match="failed to strictify tool parameter schema"):
        OAuthCodexStateProvider(instructions="x", client=fake_client).validate_tool_schemas(
            ToolRegistry.from_callables([consume_recursive]).schemas()
        )

    with pytest.raises(LLMOutputError, match="failed to strictify tool parameter schema"):
        Agent(
            instructions="x",
            state_models={"reasoning": "gpt-5.3-codex"},
            tools=[consume_recursive],
        )



from __future__ import annotations

import pytest
from pydantic import BaseModel

from fabrix.agent import _DEFAULT_MAX_STEPS, Agent
from fabrix.errors import RetryableLLMOutputError
from fabrix.events import ReasoningEvent, ResponseEvent, TaskFailedEvent, ToolEvent
from fabrix.graph.state import (
    NextState,
    ReasoningState,
    ResponseState,
    StateEnvelope,
    ToolCallState,
)
from fabrix.llm.oauth_codex import DEFAULT_MODEL, OAuthCodexStateProvider
from fabrix.messages import ImageMessage, TextMessage
from fabrix.tools import ToolOutput

_STATE_MODELS = {
    "reasoning": "gpt-5.3-codex",
    "tool_call": "gpt-5.3-codex",
    "response": "gpt-5.3-codex",
}


class AddInput(BaseModel):
    a: int
    b: int


def add_numbers(payload: AddInput) -> ToolOutput:
    return ToolOutput.json({"sum": payload.a + payload.b})


class DoubleInput(BaseModel):
    value: int


async def slow_double(payload: DoubleInput) -> ToolOutput:
    return ToolOutput.json({"value": payload.value * 2})


def _text_message(text: str) -> TextMessage:
    return TextMessage(role="user", text=text)


def _patch_provider_states(monkeypatch: pytest.MonkeyPatch, states: list[object]) -> None:
    iterator = iter(states)

    async def fake_generate_state(self, **_):  # type: ignore[no-untyped-def]
        return StateEnvelope(state=next(iterator))

    monkeypatch.setattr(OAuthCodexStateProvider, "generate_state", fake_generate_state)


def _patch_provider_loop(monkeypatch: pytest.MonkeyPatch) -> None:
    async def fake_generate_state(self, **_):  # type: ignore[no-untyped-def]
        return StateEnvelope(
            state=ReasoningState(
                next_state=NextState.reasoning,
                reasoning="keep working",
                focus="iteration",
            )
        )

    monkeypatch.setattr(OAuthCodexStateProvider, "generate_state", fake_generate_state)


def test_agent_rejects_removed_init_parameters() -> None:
    with pytest.raises(TypeError):
        Agent(
            instructions="x",
            state_models=_STATE_MODELS,
            tools=[add_numbers],
            max_steps=1,  # type: ignore[call-arg]
        )


def test_agent_rejects_removed_model_parameter() -> None:
    with pytest.raises(TypeError):
        Agent(
            instructions="x",
            model="gpt-5.3-codex",  # type: ignore[call-arg]
            tools=[add_numbers],
        )


def test_agent_accepts_instruction_callable(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def fake_provider_init(self: OAuthCodexStateProvider, **kwargs: object) -> None:
        captured["instructions"] = kwargs["instructions"]
        captured["default_model"] = kwargs["default_model"]
        captured["state_models"] = kwargs["state_models"]

    def fake_validate_tool_schemas(
        self: OAuthCodexStateProvider, tool_schemas: list[dict[str, object]]
    ) -> None:
        del self, tool_schemas

    monkeypatch.setattr(OAuthCodexStateProvider, "__init__", fake_provider_init)
    monkeypatch.setattr(OAuthCodexStateProvider, "validate_tool_schemas", fake_validate_tool_schemas)

    Agent(
        instructions=lambda: "Use strict policies.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    instruction_source = captured["instructions"]
    assert captured["default_model"] == DEFAULT_MODEL
    assert captured["state_models"] == _STATE_MODELS
    assert callable(instruction_source)
    assert instruction_source() == "Use strict policies."


def test_agent_forwards_custom_default_model(monkeypatch: pytest.MonkeyPatch) -> None:
    captured: dict[str, object] = {}

    def fake_provider_init(self: OAuthCodexStateProvider, **kwargs: object) -> None:
        captured["default_model"] = kwargs["default_model"]

    def fake_validate_tool_schemas(
        self: OAuthCodexStateProvider, tool_schemas: list[dict[str, object]]
    ) -> None:
        del self, tool_schemas

    monkeypatch.setattr(OAuthCodexStateProvider, "__init__", fake_provider_init)
    monkeypatch.setattr(OAuthCodexStateProvider, "validate_tool_schemas", fake_validate_tool_schemas)

    Agent(
        instructions="x",
        default_model="gpt-5.3-codex-custom",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    assert captured["default_model"] == "gpt-5.3-codex-custom"


def test_agent_rejects_non_string_non_callable_instructions() -> None:
    with pytest.raises(TypeError, match="instructions must be a string or a callable returning a string"):
        Agent(
            instructions=123,  # type: ignore[arg-type]
            state_models=_STATE_MODELS,
            tools=[add_numbers],
        )


@pytest.mark.asyncio
async def test_stream_rejects_when_messages_missing() -> None:
    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    with pytest.raises(ValueError, match="messages must be a non-empty list of TextMessage/ImageMessage objects"):
        async for _event in agent.run_stream(messages=[]):
            pass


@pytest.mark.asyncio
async def test_stream_rejects_dict_messages() -> None:
    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    with pytest.raises(TypeError, match="messages must be a list of TextMessage/ImageMessage objects"):
        async for _event in agent.run_stream(messages=[{"role": "user", "content": "x"}]):  # type: ignore[list-item]
            pass


@pytest.mark.asyncio
async def test_stream_accepts_image_only_input(monkeypatch: pytest.MonkeyPatch) -> None:
    _patch_provider_states(
        monkeypatch,
        [
            ReasoningState(
                next_state=NextState.response,
                reasoning="Analyze the image first",
                focus="visual inspection",
            ),
            ResponseState(
                next_state=None,
                response="image analyzed",
                audience="user",
            ),
        ],
    )

    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    messages = [ImageMessage(role="user", image="https://example.com/input.png")]
    events = [event async for event in agent.run_stream(messages=messages)]

    assert isinstance(events[-1], ResponseEvent)
    assert events[-1].response == "image analyzed"


@pytest.mark.asyncio
async def test_stream_allows_empty_response_payload(monkeypatch: pytest.MonkeyPatch) -> None:
    _patch_provider_states(
        monkeypatch,
        [
            ReasoningState(
                next_state=NextState.response,
                reasoning="No content needed",
                focus="emit empty response",
            ),
            ResponseState(
                next_state=None,
                response=None,
                parts=None,
                audience="user",
            ),
        ],
    )

    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("empty response")])]
    assert isinstance(events[-1], ResponseEvent)
    assert events[-1].response is None
    assert events[-1].parts is None


@pytest.mark.asyncio
async def test_stream_preserves_response_parts_in_history(monkeypatch: pytest.MonkeyPatch) -> None:
    async def fake_generate_state(self, **kwargs):  # type: ignore[no-untyped-def]
        step = kwargs["step"]
        if step == 1:
            return StateEnvelope(
                state=ReasoningState(
                    next_state=NextState.response,
                    reasoning="Respond with an image part",
                    focus="emit multimodal response",
                )
            )
        if step == 2:
            return StateEnvelope(
                state=ResponseState(
                    next_state=NextState.reasoning,
                    response=None,
                    parts=[{"type": "image", "image_url": "https://example.com/result.png"}],
                    audience="user",
                )
            )
        if step == 3:
            history = kwargs["history"]
            response_items = [item for item in history if item.get("kind") == "response"]
            assert response_items
            assert response_items[-1]["response"] is None
            assert response_items[-1]["parts"] == [
                {"type": "image", "image_url": "https://example.com/result.png", "caption": None}
            ]
            return StateEnvelope(
                state=ReasoningState(
                    next_state=NextState.response,
                    reasoning="Wrap up after image response",
                    focus="final answer",
                )
            )
        if step == 4:
            return StateEnvelope(
                state=ResponseState(
                    next_state=None,
                    response="done",
                    audience="user",
                )
            )

        raise AssertionError(f"unexpected step: {step}")

    monkeypatch.setattr(OAuthCodexStateProvider, "generate_state", fake_generate_state)

    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("response parts")])]
    response_events = [event for event in events if isinstance(event, ResponseEvent)]
    assert len(response_events) == 2
    assert response_events[0].response is None
    assert response_events[0].parts is not None
    assert response_events[0].parts[0].type == "image"
    assert response_events[-1].response == "done"


@pytest.mark.asyncio
async def test_stream_forwards_messages_to_state_provider(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: list[dict[str, object]] = []

    async def fake_generate_state(self, **kwargs):  # type: ignore[no-untyped-def]
        calls.append(kwargs)
        if kwargs["current_state"] == NextState.reasoning:
            return StateEnvelope(
                state=ReasoningState(
                    next_state=NextState.response,
                    reasoning="ready",
                    focus="respond now",
                )
            )
        return StateEnvelope(
            state=ResponseState(
                next_state=None,
                response="done",
                audience="user",
            )
        )

    monkeypatch.setattr(OAuthCodexStateProvider, "generate_state", fake_generate_state)

    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    messages = [_text_message("hello")]
    events = [event async for event in agent.run_stream(messages=messages)]

    assert isinstance(events[-1], ResponseEvent)
    assert calls
    assert calls[0]["messages"] == messages


@pytest.mark.asyncio
async def test_response_none_terminates_without_extra_state_call(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    call_count = 0

    async def fake_generate_state(self, **kwargs):  # type: ignore[no-untyped-def]
        nonlocal call_count
        call_count += 1
        if kwargs["current_state"] == NextState.reasoning:
            return StateEnvelope(
                state=ReasoningState(
                    next_state=NextState.response,
                    reasoning="answering now",
                    focus="final response",
                )
            )
        if kwargs["current_state"] == NextState.response:
            return StateEnvelope(
                state=ResponseState(
                    next_state=None,
                    response="final answer",
                    audience="user",
                )
            )
        raise AssertionError("generate_state must not be called after terminal response")

    monkeypatch.setattr(OAuthCodexStateProvider, "generate_state", fake_generate_state)

    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("end now")])]

    assert call_count == 2
    assert isinstance(events[-1], ResponseEvent)
    assert events[-1].response == "final answer"


@pytest.mark.asyncio
async def test_stream_emits_events_in_expected_order(monkeypatch: pytest.MonkeyPatch) -> None:
    _patch_provider_states(
        monkeypatch,
        [
            ReasoningState(
                next_state=NextState.tool_call,
                reasoning="Need calculator",
                focus="compute",
            ),
            ToolCallState(
                next_state=NextState.reasoning,
                tool_calls=[{"name": "add_numbers", "arguments": {"a": 2, "b": 5}}],
            ),
            ReasoningState(
                next_state=NextState.response,
                reasoning="Now I can answer",
                focus="user response",
            ),
            ResponseState(
                next_state=None,
                response="The sum is 7.",
                audience="user",
            ),
        ],
    )

    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("add 2 and 5")])]

    assert any(isinstance(event, ReasoningEvent) for event in events)
    assert any(isinstance(event, ResponseEvent) for event in events)
    tool_events = [event for event in events if isinstance(event, ToolEvent)]
    assert [event.phase for event in tool_events] == ["start", "finish"]
    assert tool_events[-1].result is not None
    assert tool_events[-1].result.ok is True
    assert isinstance(tool_events[-1].result.output, ToolOutput)
    assert isinstance(events[-1], ResponseEvent)
    assert events[-1].response == "The sum is 7."


@pytest.mark.asyncio
async def test_stream_executes_multiple_tools_sequentially(monkeypatch: pytest.MonkeyPatch) -> None:
    _patch_provider_states(
        monkeypatch,
        [
            ReasoningState(
                next_state=NextState.tool_call,
                reasoning="Need two calls",
                focus="sequential",
            ),
            ToolCallState(
                next_state=NextState.tool_call,
                tool_calls=[{"name": "slow_double", "arguments": {"value": 2}}],
            ),
            ToolCallState(
                next_state=NextState.response,
                tool_calls=[{"name": "slow_double", "arguments": {"value": 3}}],
            ),
            ResponseState(
                next_state=None,
                response="done",
                audience="user",
            ),
        ],
    )

    agent = Agent(
        instructions="Follow graph.",
        state_models=_STATE_MODELS,
        tools=[slow_double],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("run tools")])]

    tool_events = [event for event in events if isinstance(event, ToolEvent)]
    assert [event.phase for event in tool_events] == ["start", "finish", "start", "finish"]
    for event in tool_events:
        if event.phase == "finish":
            assert event.result is not None
            assert event.result.ok is True
    assert isinstance(events[-1], ResponseEvent)


@pytest.mark.asyncio
async def test_all_to_all_state_flow(monkeypatch: pytest.MonkeyPatch) -> None:
    _patch_provider_states(
        monkeypatch,
        [
            ReasoningState(
                next_state=NextState.tool_call,
                reasoning="Route to tool call",
                focus="transition 1",
            ),
            ToolCallState(
                next_state=NextState.response,
                tool_calls=[{"name": "add_numbers", "arguments": {"a": 1, "b": 1}}],
            ),
            ResponseState(
                next_state=NextState.reasoning,
                response="intermediate",
                audience="user",
            ),
            ReasoningState(
                next_state=NextState.response,
                reasoning="switch to response",
                focus="transition 4",
            ),
            ResponseState(
                next_state=NextState.tool_call,
                response="one more tool",
                audience="user",
            ),
            ToolCallState(
                next_state=NextState.response,
                tool_calls=[{"name": "add_numbers", "arguments": {"a": 2, "b": 3}}],
            ),
            ResponseState(
                next_state=None,
                response="final",
                audience="user",
            ),
        ],
    )

    agent = Agent(
        instructions="Follow graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("all transitions")])]
    assert isinstance(events[-1], ResponseEvent)
    assert events[-1].response == "final"


@pytest.mark.asyncio
async def test_tool_errors_are_emitted(monkeypatch: pytest.MonkeyPatch) -> None:
    _patch_provider_states(
        monkeypatch,
        [
            ReasoningState(
                next_state=NextState.tool_call,
                reasoning="Try bad tool calls",
                focus="validation",
            ),
            ToolCallState(
                next_state=NextState.response,
                tool_calls=[{"name": "missing_tool", "arguments": {"a": 1}}],
            ),
            ResponseState(
                next_state=None,
                response="done",
                audience="user",
            ),
        ],
    )

    agent = Agent(
        instructions="Follow the graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("tool errors")])]
    finished_tools = [
        event for event in events if isinstance(event, ToolEvent) and event.phase == "finish"
    ]
    assert len(finished_tools) == 1
    assert finished_tools[0].result is not None
    assert finished_tools[0].result.ok is False
    assert finished_tools[0].result.error == "tool not found: missing_tool"


@pytest.mark.asyncio
async def test_max_steps_without_response_emits_no_terminal_event(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _patch_provider_loop(monkeypatch)

    agent = Agent(
        instructions="Keep reasoning.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("loop")])]
    assert len(events) == _DEFAULT_MAX_STEPS
    assert isinstance(events[-1], ReasoningEvent)
    assert not any(isinstance(event, TaskFailedEvent) for event in events)


@pytest.mark.asyncio
async def test_retryable_llm_errors_are_retried(monkeypatch: pytest.MonkeyPatch) -> None:
    call_count = 0

    async def fake_generate_state(self, **kwargs):  # type: ignore[no-untyped-def]
        nonlocal call_count
        call_count += 1
        if kwargs["current_state"] == NextState.reasoning and call_count == 1:
            raise RetryableLLMOutputError("retry once")
        if kwargs["current_state"] == NextState.reasoning:
            return StateEnvelope(
                state=ReasoningState(
                    next_state=NextState.response,
                    reasoning="ready",
                    focus="answer",
                )
            )
        return StateEnvelope(
            state=ResponseState(
                next_state=None,
                response="done",
                audience="user",
            )
        )

    monkeypatch.setattr(OAuthCodexStateProvider, "generate_state", fake_generate_state)

    agent = Agent(
        instructions="Follow graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("retry")])]
    assert isinstance(events[-1], ResponseEvent)
    assert events[-1].response == "done"
    assert call_count == 3


@pytest.mark.asyncio
async def test_retryable_llm_errors_fail_after_retries(monkeypatch: pytest.MonkeyPatch) -> None:
    call_count = 0

    async def fake_generate_state(self, **_):  # type: ignore[no-untyped-def]
        nonlocal call_count
        call_count += 1
        raise RetryableLLMOutputError("always bad")

    monkeypatch.setattr(OAuthCodexStateProvider, "generate_state", fake_generate_state)

    agent = Agent(
        instructions="Follow graph.",
        state_models=_STATE_MODELS,
        tools=[add_numbers],
    )

    events = [event async for event in agent.run_stream(messages=[_text_message("retry fail")])]
    assert isinstance(events[-1], TaskFailedEvent)
    assert events[-1].error_code == "llm_error"
    assert "retry attempts exhausted" in events[-1].message
    assert call_count == 5
